#Project template for CHE210D

#import python modules
import numpy as np
import time

#import compiled fortran library
import ex4lib

import sys
import os
import math
import atomwrite
#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol, L,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [ex4lib.calcenergy(Pos + NormDir * x, M, L, Cut) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = ex4lib.calcenergy(Pos + NormDir * dx, M, L, Cut)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [ex4lib.calcenergy(Pos + NormDir * 2. * dx, M, L, Cut)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [ex4lib.calcenergy(Pos + NormDir * Dists[-1], M, L, Cut)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = ex4lib.calcenergy(Pos + NormDir * d3, M, L, Cut)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = ex4lib.calcenergy(Pos + NormDir * d3, M, L, Cut)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG, L):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, Cut
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = ex4lib.calcenergyforces(Pos, M, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS, L)
        OldForces = Forces.copy()
        PE, Forces = ex4lib.calcenergyforces(Pos, M, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos, L):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = ex4lib.calcenergyforces(Pos, M, L, Cut, Forces)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

#### YOUR CODE HERE ####
def MCdisP(PEnergy, W, N, Pos, dr, Nreject, T, M, L):
    MoveAtom = np.random.randint(len(Pos))
    OldPos = Pos[MoveAtom,:].copy()
    OldPEnergy, OldW = ex4lib.calcenergyvirial(Pos, M, L, Cut, MoveAtom)
    
    Pos[MoveAtom,0] = OldPos[0] + np.random.uniform(-dr,dr)
    Pos[MoveAtom,1] = OldPos[1] + np.random.uniform(-dr,dr)
    Pos[MoveAtom,2] = OldPos[2] + np.random.uniform(-dr,dr)
    
    NewPEnergy, NewW = ex4lib.calcenergyvirial(Pos, M, L, Cut, MoveAtom)   
    
    dU = NewPEnergy-OldPEnergy
    dW = NewW-OldW
    
    if dU < 0:
        Reject = False
        PEnergy += dU
        W += dW
    else:
        if math.exp(-dU/T) > np.random.uniform(0.0, 1.0):
            Reject = False
            PEnergy += dU
            W += dW
        else:
            Reject = True
                    
    
    if Reject:
        Pos[MoveAtom,:] = OldPos
        Nreject+=1

    return Pos, PEnergy, W, Nreject

def MCdisV(PEnergy1, N1, Pos1, L1, NrejectV1, PEnergy2, N2, Pos2, L2, NrejectV2, dr, T, M, dv):
    dV1 = np.random.uniform(0,dv)
    V1 = L1**(3.) 
    V2 = L2**(3.)    
        
    BoxN = np.random.randint(2)    
    AcceptV = True
    
    if BoxN == 1:
        newV1 = V1 + dV1
        newV2 = V2 - dV1
        if newV2 < 0:
            newV2 = 0
            AcceptV = False
            Reject = True
        
        dL1 = newV1**(1/3.)
        dL2 = newV2**(1/3.)
    else:
        newV1 = V1 - dV1
        newV2 = V2 + dV1
        
        if newV1 < 0:
            newV1 = 0
            AcceptV = False
            Reject = True
            
        dL1 = newV1**(1/3.)
        dL2 = newV2**(1/3.)
        
    OldPos1 = Pos1.copy()
    OldPos2 = Pos2.copy()
    
    OldPEnergy1 = ex4lib.calcenergy(Pos1, M, L1, Cut)
    OldPEnergy2 = ex4lib.calcenergy(Pos2, M, L2, Cut)
    
    Pos1[:,:] = OldPos1[:,:]*dL1/L1 
    Pos2[:,:] = OldPos2[:,:]*dL2/L2
    
    NewPEnergy1 = ex4lib.calcenergy(Pos1, M, dL1, Cut)
    NewPEnergy2 = ex4lib.calcenergy(Pos2, M, dL2, Cut) 
    
    dU1 = NewPEnergy1-OldPEnergy1
    dU2 = NewPEnergy2-OldPEnergy2
    
    if BoxN ==1 and AcceptV:
        if ((V1+dV1)/V1)*(N1)*((V2-dV1)/V2)**(N2)*math.exp(-dU1-dU2) > np.random.uniform(0.0, 1.0):
            Reject = False
            PEnergy1 += dU1
            PEnergy2 += dU2
            V1 = newV1
            V2 = newV2
            L1 = V1**(1/3.)
            L2 = V2**(1/3.)            
        else:
            Reject = True
    else:
        if ((V1-dV1)/V1)*(N1)*((V2+dV1)/V2)**(N2)*math.exp(-dU1-dU2) > np.random.uniform(0.0, 1.0):
            Reject = False
            PEnergy1 += dU1
            PEnergy2 += dU2
            V1 = newV1
            V2 = newV2
            L1 = V1**(1/3.)
            L2 = V2**(1/3.)             
        else:
            Reject = True
    
    if Reject:
        Pos1 = OldPos1
        Pos2 = OldPos2
        NrejectV1+=1
        NrejectV2+=1
    
    
    return Pos1, PEnergy1, NrejectV1, L1, Pos2, PEnergy2, NrejectV2, L2

def MCmoveP(PEnergy1, N1, Pos1, L1, NrejectP1, PEnergy2, N2, Pos2, L2, NrejectP2, dr, T, M, dv):
    
    V1 = L1**(3.) 
    V2 = L2**(3.)
    
    OldPos1 = Pos1.copy()
    OldPos2 = Pos2.copy()
    
    OldPEnergy1 = ex4lib.calcenergy(OldPos1, M, L1, Cut)
    OldPEnergy2 = ex4lib.calcenergy(OldPos2, M, L2, Cut)    
    
    
    BoxN = np.random.randint(2)
    
    if BoxN == 1:    
        deleteAtom = np.random.randint(len(Pos1))
        Pos1 = np.zeros((len(OldPos1)-1,3), float)    
        for k in range(len(OldPos1)-1):
            if k < deleteAtom:
                Pos1[k,:] = OldPos1[k,:]
            else:
                Pos1[k,:] = OldPos1[k+1,:]
        
        Pos2 = np.zeros((len(OldPos2)+1,3), float)       
        
        for i in range(len(OldPos2)):
            Pos2[i, :]= OldPos2[i,:]
        
        x = np.random.uniform(0,L2)
        y = np.random.uniform(0,L2)
        z = np.random.uniform(0,L2)
        
        Pos2[len(OldPos2), 0] = x
        Pos2[len(OldPos2), 1] = y
        Pos2[len(OldPos2), 2] = z
           
    else:
        deleteAtom = np.random.randint(len(Pos2))        
        
        Pos2 = np.zeros((len(OldPos2)-1,3), float)    
        for k in range(len(OldPos2)-1):
            if k < deleteAtom:
                Pos2[k,:] = OldPos2[k,:]
            else:
                Pos2[k,:] = OldPos2[k+1,:]
        
        Pos1 = np.zeros((len(OldPos1)+1,3), float)       
        for i in range(len(OldPos1)):
            Pos1[i, :]= OldPos1[i,:]
        
        x = np.random.uniform(0,L1)
        y = np.random.uniform(0,L1)
        z = np.random.uniform(0,L1)
        
        Pos1[len(OldPos1), 0] = x
        Pos1[len(OldPos1), 1] = y
        Pos1[len(OldPos1), 2] = z
    
    
        
    NewPEnergy1 = ex4lib.calcenergy(Pos1, M, L1, Cut)
    NewPEnergy2 = ex4lib.calcenergy(Pos2, M, L2, Cut) 
    
    dU1 = NewPEnergy1-OldPEnergy1
    dU2 = NewPEnergy2-OldPEnergy2
    
        
    if BoxN == 1:
        if N1*V2/(N2+1)/V1*math.exp(-dU1-dU2) > np.random.uniform(0.0, 1.0):
            Reject = False
            PEnergy1 += dU1
            PEnergy2 += dU2
            N1 = len(Pos1)
            N2 = len(Pos2)                
        else:
            Reject = True
        
    else:        
        if N2*V1/(N1+1)/V2*math.exp(-dU1-dU2) > np.random.uniform(0.0, 1.0):
            Reject = False
            PEnergy1 += dU1
            PEnergy2 += dU2
            N1 = len(Pos1)
            N2 = len(Pos2)                
        else:
            Reject = True  
    
    if N1 == 0 or N2 == 0:
        Reject = True    
    
    if Reject:
        Pos1 = np.zeros((len(OldPos1),3), float)       
        for i in range(len(OldPos1)):
            Pos1[i, :]= OldPos1[i,:]
            
        Pos2 = np.zeros((len(OldPos2),3), float)       
        for j in range(len(OldPos2)):
            Pos2[j, :]= OldPos2[j,:]
        
        N1 = len(Pos1)
        N2 = len(Pos2)
        NrejectP1+= 1
        NrejectP2+= 1

    return Pos1, PEnergy1, NrejectP1, N1, Pos2, PEnergy2, NrejectP2, N2
    
def TestProject():
    global Cut, M
    T = 1.00
    dx = 0.001
    dt = 0.001
    Dim = 3
    rc = Cut
    EFracTolLS = 10**(-8)
    EFracTolCG = 10**(-10)
    
    
    rho = 0.3
    
    M = 1
    NT = 240
    NStepsEquil1 = 1000*NT
    NStepsProd = 500*NT    
    FullFreq = 50
    PBDFreq = 100
    
       
    dr=0.2
    Pdb1 = atomwrite.pdbfile("anim1.pdb", Compressed = False)
    #Pdb2 = atomwrite.pdbfile("anim2.pdb", Compressed = False)      
    wf = file("C:\\Python27\\Che210D\\movie.txt", "w")   
    
    for trial in (range(1)):
        N1 = 120 
        N2 = 120
        dv = (N1/rho)/5.    
    
        L1 = (N1/rho)**(1/3.)
        V1 = L1**(3.)
    
        L2 = (N2/rho)**(1/3.)
        V2 = L2**(3.)
        
        Nreject1 = 0.
        Nreject2 = 0.
        NrejectV1 = 0.
        NrejectV2 = 0.
        NrejectP1 = 0.
        NrejectP2 = 0.
        
        rho1Lis=[]
        rho2Lis=[]
        
        Pos1 = InitPositions(N1, L1)
        PEnergy1, Pos1 = ConjugateGradient(Pos1, dx, EFracTolLS, EFracTolCG, L1)
        PEnergy1, W1 = ex4lib.calcenergyvirialtotal(Pos1, M, L1, rc)
        
        Pos2 = InitPositions(N2, L2)
        PEnergy2, Pos2 = ConjugateGradient(Pos2, dx, EFracTolLS, EFracTolCG, L2)
        PEnergy2, W2 = ex4lib.calcenergyvirialtotal(Pos2, M, L2, rc)
        
        for i in (range(NStepsEquil1)):
            moveW = np.random.randint(4)
            
            if moveW == 1:
                Pos1, Penergy1, W1, Nreject1 = MCdisP(PEnergy1, W1, N1, Pos1, dr, Nreject1, T, M, L1)
            elif moveW ==2:
                Pos2, Penergy2, W2, Nreject2 = MCdisP(PEnergy2, W2, N2, Pos2, dr, Nreject2, T, M, L2)
            elif moveW ==3:
                Pos1, PEnergy1, NrejectV1, L1, Pos2, PEnergy2, NrejectV2, L2 = MCdisV(PEnergy1, N1, Pos1, L1, NrejectV1, PEnergy2, N2, Pos2, L2, NrejectV2, dr, T, M, dv)
            else:
                Pos1, PEnergy1, NrejectP1, N1, Pos2, PEnergy2, NrejectP2, N2 = MCmoveP(PEnergy1, N1, Pos1, L1, NrejectP1, PEnergy2, N2, Pos2, L2, NrejectP2, dr, T, M, dv)
            
            #print(PEnergy1, PEnergy2, N1, N2, N1/L1**(3.), N2/L2**(3.),(N1+N2)/(L1**(3.)+L2**(3.)))
            if (i % FullFreq == 0):
                PEnergy1, W1 = ex4lib.calcenergyvirialtotal(Pos1, M, L1, rc)
                PEnergy2, W2 = ex4lib.calcenergyvirialtotal(Pos2, M, L2, rc)
    #        if (i % 100 == 0):
    #            wf.write("%i\t %.5f\t\n" %(i/N, PEnergy))
        
       
        print(float(Nreject1/NStepsEquil1))
        print(float(Nreject2/NStepsEquil1))
        print(float(NrejectV1/NStepsEquil1))
        print(float(NrejectV2/NStepsEquil1))
        print(float(NrejectP1/NStepsEquil1))
        print(float(NrejectP2/NStepsEquil1))
                    
        for j in (range(NStepsProd)):
            moveW = np.random.randint(4)
            
            if moveW == 1:
                Pos1, Penergy1, W1, Nreject1 = MCdisP(PEnergy1, W1, N1, Pos1, dr, Nreject1, T, M, L1)
            elif moveW ==2:
                Pos2, Penergy2, W2, Nreject2 = MCdisP(PEnergy2, W2, N2, Pos2, dr, Nreject2, T, M, L2)
            elif moveW ==3:
                Pos1, PEnergy1, NrejectV1, L1, Pos2, PEnergy2, NrejectV2, L2 = MCdisV(PEnergy1, N1, Pos1, L1, NrejectV1, PEnergy2, N2, Pos2, L2, NrejectV2, dr, T, M, dv)
            else:
                Pos1, PEnergy1, NrejectP1, N1, Pos2, PEnergy2, NrejectP2, N2 = MCmoveP(PEnergy1, N1, Pos1, L1, NrejectP1, PEnergy2, N2, Pos2, L2, NrejectP2, dr, T, M, dv)
            
            #print(PEnergy1, PEnergy2, N1, N2, N1/L1**(3.), N2/L2**(3.),(N1+N2)/(L1**(3.)+L2**(3.)))
            if (j % FullFreq == 0):
                PEnergy1, W1 = ex4lib.calcenergyvirialtotal(Pos1, M, L1, rc) 
                PEnergy2, W2 = ex4lib.calcenergyvirialtotal(Pos2, M, L2, rc)
            rho1Lis.append(N1/L1**(3.))
            rho2Lis.append(N2/L2**(3.))
            
            if(j % PBDFreq == 0):
                PosT = np.zeros((NT,3), float)       
                for ai in range(N1):
                    PosT[ai, :]= Pos1[ai,:]
            
                for aj in range(N2):
                    PosT[N1+aj, 0]= Pos2[aj,0] + 15
                    PosT[N1+aj, 1]= Pos2[aj,1]
                    PosT[N1+aj, 2]= Pos2[aj,2]
                Pdb1.write(PosT)
                #Pdb2.write(Pos2)
            
            #if (j % WFreq == 0):
                #wf.write("%.5f \t %.5f \t %.5f \t %.5f\n" %(N1, N2, L1**(3.), L2**(3.)))
                 
                       
        rho1avg = sum(rho1Lis)/float(len(rho1Lis))
        rho2avg = sum(rho2Lis)/float(len(rho2Lis))    
        
        NrejectPer1 = float(float(Nreject1)/float(NStepsProd))
        NrejectPer2 = float(float(Nreject2)/float(NStepsProd))
        #P1 = N1*T/V1-W1avg/3./V1
        #P2 = N2*T/V2-W2avg/3./V2
        
    #    print(P,Wavg, NrejectPer)
        print(rho1avg, rho2avg, (N1+N2)/(L1**(3.)+L2**(3.)))
        wf.write("%.5f\t%.5f\t%.5f\t%.5f\t%.5f\n" %(rho1avg, rho2avg, N1 , N2, (N1+N2)/(L1**(3.)+L2**(3.))))
    
    wf.close()        
    Pdb1.close()
    
    
#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####

